% [X,E,G,a] = Opensoftmax(xs,kt,l,A,b,[x0,tol,maxit,rho])
%
% Optimisation by Newton's method of K-class softmax with penalty |x-xs|:
%   min_x{ E(x) = l.|x-xs| - log(fk) }
% where fk = e^zk/sum_j{e^zj}; z = A.x+b; x,xs of Dx1; A of KxD; b of Kx1; l>0.
% Optimising E means finding an x that is close to a given xs (the source
% instance) while maximising the probability of class k (the target class).
%
% This function also handles equality constraints where several variables are
% set to constant values. Let I be a set of indices (for the fixed variables)
% and y be a vector with the corresponding constant values. For example,
% I = [2 5] and y = [2.3;-1] means that x2 = 2.3 and x5 = -1. Then the
% solution x is obtained as follows:
%   J = setdiff(1:length(xs),I);
%   [X,E,G,a] = Opensoftmax(xs(J),kt,l,A(:,J),A(:,I)*y+b);
%   x = xs; x(I) = y; x(J) = X(end,:)';
%
% In:
%   xs: Dx1 vector, the source instance.
%   kt: integer in 1:K, the target class.
%   l: positive scalar, penalty parameter.
%   A: KxD matrix of softmax vectors.
%   b: Kx1 vector of softmax biases.
%   x0: Dx1 vector, the starting point. Default: xs.
%   tol: tolerance to stop iterating (minimum gradient norm). Default: 1e-6.
%   maxit: maximum number of iterations. Default: 300.
%   rho: scalar in (0,1), backtracking l.s. decrease factor. Default: 0.8.
% Out (where I = number of iterations run, first entry is for x0):
%   X: (I+1)xD list of iterates.
%   E: (I+1)x1 list of the objfcn value at the iterates.
%   G: (I+1)x1 list of the objfcn gradient norm at the iterates.
%   a: (I+1)x1 list of step size values.
%
% Notes:
% - Underflow can happen in function softmaxf(x,A,b), in particular if A and b
%   are such that some class regions are empty (so the corresponding softmax
%   value is 0). This affects the following operations, but seems to work
%   correctly anyway:
%   . "(AA + diag(l./f)) \ ...": Matlab shows a warning about a matrix being
%     (nearly) singular.
%   . "-log(f(kt))": this becomes Inf and makes the objfcn Inf.
% - If tol is very small (around 1e-7, but it depends on the case) then the
%   computer's floating point arithmetic (double precision) makes it impossible
%   to lower norm(g) below tol. Specifically, the backtracking l.s. fails to
%   find a step size that lowers the objfcn. In that case, we simply stop and
%   return the current iterate. We do this if the step size falls below 1e-5.
%   Generally, we can't expect to reduce the gradient norm below sqrt(eps)
%   (around 1e-8 in double precision). See e.g. Gill et al. "Practical
%   optimization", Academic Press, 1981, p. 301ff.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2020 by Miguel A. Carreira-Perpinan

function [X,E,G,a] = Opensoftmax(xs,kt,l,A,b,x0,tol,maxit,rho)

% warning('off','MATLAB:nearlySingularMatrix');
% warning('off','MATLAB:singularMatrix');

% ---------- Argument defaults ----------
if ~exist('x0','var') || isempty(x0) x0 = xs; end;
if ~exist('tol','var') || isempty(tol) tol = 1e-6; end;
if ~exist('maxit','var') || isempty(maxit) maxit = 300; end;
if ~exist('rho','var') || isempty(rho) rho = 0.8; end;
% ---------- End of "argument defaults" ----------

Ak = bsxfun(@minus,A,A(kt,:)); bb = b - b(kt); AA = Ak*Ak';

k = 0;				% Number of iterations
x = x0;				% Current iterate
f = softmaxf(x,A,b); v = Ak'*f; xxs = x - xs; g = l*xxs + v;	% g = gradient
e = sum((xxs).^2)*l/2 - log(f(kt));				% e = objfcn
X = x'; E = e; G = norm(g); a = NaN;	% Lists of results

while (G(end) > tol) & (k < maxit)

  % New search direction p = -inv(Hessian)*gradient.
  %HH = l*eye(D,D) + Ak'*(diag(f)-f*f')*Ak; p = -HH\g; % DxD Hessian, slow
  U = Ak'*((AA + diag(l./f)) \ (Ak*[v xxs]));
  iHv = (v - U(:,1))/l; c = 1/(1-v'*iHv);
  p = - ( (c*(l*(iHv'*xxs)+1))*iHv + xxs - U(:,2) );

  % Backtracking line search: decrease step size alpha until objfcn decreases
  alpha = 1; f = softmaxf(x+p,A,b); enew = sum((xxs+p).^2)*l/2 - log(f(kt));
  while enew > e
    if alpha < 1e-5		% safeguard: precision loss, stop iterating
      disp('Precision loss, backtracking stopped with step size < 1e-5');
      return;
    end;
    alpha = rho*alpha;
    f = softmaxf(x+alpha*p,A,b); enew = sum((xxs+alpha*p).^2)*l/2 - log(f(kt));
  end

  x = x + alpha*p; e = enew;			% New iterate, objfcn value...
  v = Ak'*f; xxs = x - xs; g = l*xxs + v;	% ...and gradient
  X = [X;x']; E = [E;e]; G = [G;norm(g)]; a = [a;alpha];	% Update lists
  k = k + 1;
  
end

% warning('on','MATLAB:nearlySingularMatrix');
% warning('on','MATLAB:singularMatrix');


function f = softmaxf(x,A,b)

z = A*x+b; z = z - max(z); f = exp(z); f = f/sum(f);

