% function [f,g,H] = Fpensoftmax(X,W,w,v,l,o[,k]) Penalized softmax
%
% f(x) = |x-v| + l.softmax(x,W,v)
% For log-softmax, it uses log(softmax(x,W,v)).
% f has K dimensions, one per class, and sums to 1.
% We output all K dimensions by default, or a selected dimension k.
%
% In:
%   X: NxD list of row vectors.
%   W: KxD matrix.
%   w: Kx1 vector.
%   v: Dx1 vector.
%   l: nonnegative scalar.
%   o: 0 for softmax, 1 for log-softmax. Default: 0.
%   k: integer in 1:K (softmax output to select). Default: K.
% Out:
%   f: NxK list of function values. Or, if k is given, Nx1 list.
%   g: NxD list of gradient vectors (only for o=1 and given k).
%   H: DxDxN list of Hessian matrices (only for o=1 and given k, and N=1).

% Copyright (c) 2020 by Miguel A. Carreira-Perpinan

function [f,g,H] = Fpensoftmax(X,W,w,v,l,o,k)

[K,D] = size(W);
f = Fsoftmax(X,W,w);

% assumes o=1 and k is given
if nargout > 1
  V = bsxfun(@minus,W(k,:),W); g = -f*V;
  % assumes N=1
  if nargout > 2
    H = 2*eye(D,D) + V'*diag(l*f)*V - (l*g)'*g;
  end
  g = 2*bsxfun(@minus,X,v') + l*g;
end

if exist('o','var') && (o==1) f = log(f); end;
f = bsxfun(@minus,sum(bsxfun(@minus,X,v').^2,2),l*f);
if exist('k','var') && ~isempty(k) f = f(:,k); end;

