% [X,F,cgit] = Onewtoncg(f,paramf,x0[,tol,maxit,convcrit])
% Optimisation by Newton-CG
%
% Reference: p. 136ff, 139ff in:
%   Nocedal and Wright: "Numerical Optimization", Springer, 1999.
%
% Example usage & In/Out arguments: see Osteepdesc.m.
%
% In:
%   convcrit: convergence criterion for inexact Newton steps (p. 136ff);
%      one of 'linear', 'superlinear' or 'quadratic'. Default: 'quadratic'.
% Out:
%   cgit: ? x 1 list of the number of CG iterations run at each Newton step.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2006 by Miguel A. Carreira-Perpinan

function [X,F,cgit] = Onewtoncg(f,paramf,x0,tol,maxit,convcrit)

% ---------- Argument defaults ----------
if ~exist('tol','var') | isempty(tol) tol = 1e-5; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 1000; end;
if ~exist('convcrit','var') | isempty(convcrit) convcrit = 'quadratic'; end;
% ---------- End of "argument defaults" ----------

x = x0;				% Current iterate
[ff g A] = f(x',paramf{:});	% Current value of function, gradient, Hessian
X = x'; F = ff;			% Lists of iterates and their function values
cgit = NaN;			% List of numbers of CG iterations
k = 0;				% Number of iterations

while (norm(g) > tol) & (k < maxit)
  
  % Convergence tolerance for inexact Newton step with linear CG
  switch convcrit
   case 'superlinear'
    eta = min(0.5,sqrt(norm(g)));
   case 'quadratic'
    eta = min(0.5,norm(g));
   otherwise
    eta = 0.5;			% 'linear'
  end
  
  % New search direction: inexact Newton step with linear CG
  [p,kk] = Olincg1(A,-g',zeros(size(x0)),eta*norm(g));

  % Backtracking line search
  alpha = linesearch(f,paramf,x,p,ff,g);
  % No line search
% $$$   alpha = 1;
  
  x = x + alpha*p;		% New iterate
  [ff g A] = f(x',paramf{:});	% New value of function, gradient, Hessian
  X = [X;x']; F = [F;ff];	% Update lists
  cgit = [cgit;kk];
  k = k + 1;
  
end


% x = Olincg1(A,b,x0[,tol,maxit])
%
% Adaptation of "Olincg" to deal with non-positive-definite
% linear systems (as can arise in a Newton iteration).
%
% In arguments: see Osteepdesc.m.
%
% Out:
%   x: n x 1 (approximate) solution.
%   k: number of CG iterations ran.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2005 by Miguel A. Carreira-Perpinan

function [x,k] = Olincg1(A,b,x0,tol,maxit)

% ---------- Argument defaults ----------
n = length(b);
if ~exist('tol','var') | isempty(tol) tol = 1e-6; end;
if ~exist('maxit','var') | isempty(maxit) maxit = n; end;
% ---------- End of "argument defaults" ----------

x = x0;				% Current iterate
r = A*x-b;			% Current gradient of the quadratic function
p = -r;				% Current search direction
k = 0;				% Number of iterations
tmp = 1;			% Curvature indicator

while (norm(r) > tol) & (k < maxit) & (tmp > 0)
  
  tmp = p'*A*p;
  if tmp > 0			% Positive curvature
    alpha = (r'*r)/tmp;
    x = x + alpha*p;		% New iterate
    oldr = r;
    r = r + alpha*A*p;		% New gradient
    beta = (r'*r)/(oldr'*oldr);
    p = -r + beta*p;		% New search direction
    k = k + 1;
  elseif k == 0			% Non-positive curvature at first iteration
    x = -r;
  end
  
end

