% [X,LE,LI,F,CE,CI,MU,ERR,Oit,condH,G] =
%   Oqpenalty(f,paramf,E,paramE,I,paramI,O,paramO,x0[,mu0,tol,maxit])
% Constrained optimisation by quadratic-penalty method
%
% Reference: ch. 17 in:
%   Nocedal and Wright: "Numerical Optimization", 2nd ed., Springer, 2006.
%
% Example usage: the following will minimise Rosenbrock's function subject
% to 1 quadratic equality constraint and 2 quadratic inequality constraints
% (with respective parameters), using the Onewtoncg unconstrained optimiser,
% starting at the origin and running 50 iterations at most:
%
%   [X,LE,LI,F,CE,CI,MU] = Oqpenalty(@Frosenbrock,{},...
%      {@Fquad},{{A1,b1,c1}},...
%      {@Fquad,@Fquad},{{A2,b2,c2},{A3,b3,c3}},...
%      @Onewtoncg,{'quadratic'},...
%      [0;0],1,[],50);
%   figure(1);
%   fcontours(@Frosenbrock,{},...
%      {@Fquad},{{A1,b1,c1}},...
%      {@Fquad,@Fquad},{{A2,b2,c2},{A3,b3,c3}},...
%      [-2 2;-1 3],50);
%   plotseq(X);
%
% In:
%   f: a handle to a function that takes as input a list of vectors and
%      returns 3 lists: the function values, the gradients and the Hessians
%      (see Frosenbrock.m for an example in 2D).
%   paramf: cell array containing the parameters for f. Use {} to pass no
%      parameters, or to use f's default parameters.
%   E: a cell array {e1,...} of equality constraints. Each ei is a handle to
%      a function like f. Use {} if there are no equality constraints.
%   paramE: a cell array {p1,...} containing the parameters for e1,... (pi is
%      itself a cell array, like paramf). Use {} for pi to pass no parameters,
%      or to use pi's default parameters.
%   I, paramI: like E, paramE but for inequality constraints.
%   O: a handle to a function that implements an unconstrained optimisation
%      method, with the same argument format as Onewtoncg.m.
%   paramO: cell array containing the parameters for O. Use {} to pass no
%      parameters, or to use O's default parameters. Do not include x0, tol,
%      maxit (they are set by Oqpenalty).
%   x0: n x 1 vector, the starting point.
%   mu0: (positive) initial value of the penalty parameter. Default: 1.
%   tol: tolerance to stop iterating. Default: 1e-4.
%   maxit: maximum number of iterations. Default: 1000.
% Out:
%   X: ? x n list of iterates for the primal variables.
%   LE: ? x le list of estimates for the Lagrange multipliers of the equality
%      constraints.
%   LI: ? x li list of estimates for the Lagrange multipliers of the inequality
%      constraints.
%   F: ? x 1 list of the function value at the iterates.
%   CE: ? x le list of the equality constraints value at the iterates.
%   CI: ? x li list of the inequality constraints value at the iterates.
%   MU: ? x 1 list of the penalty parameter value at the iterates.
%   ERR: ? x 1 list of the error in the KKT system at the iterates.
%   Oit: ? x 1 list of the number of iterations in the unconstrained
%      optimisation method at each iterate.
%   condH: ? x 1 list of the condition number of the Hessian of the
%      quadratic-penalty function at the iterates.
%   G: ? x 1 list of the norm of the objective function gradient at the
%      iterates.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2006 by Miguel A. Carreira-Perpinan

function [X,LE,LI,F,CE,CI,MU,ERR,Oit,condH,G] = ...
      Oqpenalty(f,paramf,E,paramE,I,paramI,O,paramO,x0,mu0,tol,maxit)

% ---------- Argument defaults ----------
if ~exist('mu0','var') | isempty(mu0) mu0 = 1; end;
if ~exist('tol','var') | isempty(tol) tol = 1e-4; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 1000; end;
% ---------- End of "argument defaults" ----------

le = length(E); li = length(I);

% Method parameters
mu_rho = 0.8; tau_rho = 0.8; tau = 100*tol; maxit2 = maxit;

x = x0';			% Current iterate
mu = mu0;			% Current penalty parameter

% Current value of objective function
[ff g] = f(x,paramf{:});

% Current value of constraints
ce = zeros(1,le); ci = zeros(1,li);
for i=1:le
  tmp1 = E{i}(x,paramE{i}{:}); ce(i) = tmp1;
end
for i=1:li
  tmp1 = I{i}(x,paramI{i}{:}); ci(i) = tmp1;
end

% Lists of iterates, etc.
X = x; F = ff; CE = ce; CI = ci;
LE = repmat(NaN,1,le); LI = repmat(NaN,1,li); MU = NaN;
ERR = NaN; condH = NaN; Oit = NaN; G = norm(g);

k = 0;				% Number of iterations
notconv = 1;			% Convergence flag

while notconv
  
  % Approximate unconstrained minimisation of the penalty function
  [X1,F1] = O(@Q,{f,paramf,E,paramE,I,paramI,mu},x',tau,maxit2,paramO{:});
  x = X1(end,:);
  
% $$$   % Contours of Q and minimiser
% $$$   fcontours(@Q,{f,paramf,E,paramE,I,paramI,mu},E,paramE,I,paramI,...
% $$$             [-2 2;-1 3],200); plotseq(x); pause;

  % Estimates of Lagrange multipliers
  for i=1:le
    tmp1 = E{i}(x,paramE{i}{:}); ce(i) = tmp1;
  end
  for i=1:li
    tmp1 = I{i}(x,paramI{i}{:}); ci(i) = tmp1;
  end
  
  % Gradient of Q = gradient of Lagrangian function
  % Don't need to include norm(negpart(LI(end,:))) = 0
  [tmp1 gQ H] = Q(x,f,paramf,E,paramE,I,paramI,mu);
  err = max([norm(gQ),norm(ce),norm(negpart(ci)),norm(ci.*LI(end,:))]);

  % Update lists
  X = [X;x]; LE = [LE;-ce/mu]; LI = [LI;negpart(ci)/mu]; F = [F;F1(end)];
  CE = [CE;ce]; CI = [CI;ci]; MU = [MU;mu]; ERR = [ERR;err];
  condH = [condH;cond(H)]; Oit = [Oit;length(F1)];
  [ff g] = f(x,paramf{:}); G = [G;norm(g)];
  k = k + 1;
  
  if (err > tol) & (k < maxit)
    mun = mu_rho*mu;		% New penalty parameter
    tau = tau_rho*tau;		% New tolerance for unconstrained minimisation
% $$$     x = x;		% New starting point: current minimiser of Q
    x = Qpathtan(H,gQ,g,x,mu,mun);% New starting point: path tangent extrapol.
    mu = mun;
  else
    notconv = 0;
  end
  
end


% Xn = negpart(X) Absolute value of the negative part of X

function Xn = negpart(X)

Xn = max(-X,0);


% [ff,g,H] = Q(X,f,paramf,E,paramE,I,paramI,mu)
% Quadratic-penalty function
%
% Q(x;) = f(x) + ce(x)/2 + ([ci(x)]-)/2
%
% In:
%   X: N x n list of row vectors.
%   f, paramf, E, paramE, I, paramI: as above.
%   mu: positive number (penalty parameter).
% Out:
%   ff: N x 1 list of function values.
%   g: N x n list of gradient vectors.
%   H: n x n x N list of Hessian matrices.

% Copyright (c) 2005 by Miguel A. Carreira-Perpinan

function [ff,g,H] = Q(X,f,paramf,E,paramE,I,paramI,mu)

[N n] = size(X);

[ff g H] = f(X,paramf{:});

for i=1:length(E)
  [tmp1 tmp2 tmp3] = E{i}(X,paramE{i}{:});
  ff = ff + (tmp1.^2)/2/mu;
  g = g + (repmat(tmp1,1,n).*tmp2)/mu;
  for j=1:N
    H(:,:,j) = H(:,:,j) + (tmp2(j,:)'*tmp2(j,:) + tmp1(j)*tmp3(:,:,j))/mu;
  end
end

for i=1:length(I)
  [tmp1 tmp2 tmp3] = I{i}(X,paramI{i}{:});
  for j=find(tmp1 < 0)'
    ff(j) = ff(j) + (tmp1(j)^2)/2/mu;
    g(j,:) = g(j,:) + (tmp1(j)*tmp2(j,:))/mu;
    H(:,:,j) = H(:,:,j) + (tmp2(j,:)'*tmp2(j,:) + tmp1(j)*tmp3(:,:,j))/mu;
  end
end


% xn = Qpathtan(H,gQ,g,x,mu,mun)
%
% Extrapolation of the next starting point for the minimisation of
%   Q(x;) = f(x) + ce(x)/2 + ([ci(x)]-)/2
% given by the tangent to the path x() (the path of the minimisers of
% Q(x;) for  > 0).
%
% In:
%   H: n x n matrix, the Hessian of Q wrt x at current x, mu.
%   gQ: 1 x n vector, the gradient of Q wrt x at current x, mu.
%   g: 1 x n vector, the gradient of f wrt x at current x.
%   x: 1 x n vector, current x.
%   mu: positive nuber, current mu.
%   mun: positive nuber, new mu.
% Out:
%   xn: 1 x n vector, new x.

% Copyright (c) 2005 by Miguel A. Carreira-Perpinan

function xn = Qpathtan(H,gQ,g,x,mu,mun)

pt = H \ ((gQ-g)'/mu);		% If this linear system is singular, use xn = x
xn = x + (mun-mu)*pt';

